@@ -4,7 +4,7 @@
  * This file is part of msmtp, an SMTP client.
  *
  * Copyright (C) 2000, 2003, 2004, 2005, 2006, 2007, 2008, 2010, 2011, 2012,
- * 2014, 2015, 2016
+ * 2014, 2015, 2016, 2018, 2019
  * Martin Lambers <marlam@marlam.de>
  * Martin Stenberg <martin@gnutiken.se> (passwordeval support)
  * Scott Shumate <sshumate@austin.rr.com> (aliases support)
@@ -94,6 +94,7 @@ account_t *account_new(const char *conffile, const char *id)
     a->tls_min_dh_prime_bits = -1;
     a->tls_priorities = NULL;
     a->logfile = NULL;
+    a->logfile_time_format = NULL;
     a->syslog = NULL;
     a->aliases = NULL;
     a->proxy_host = NULL;
@@ -101,6 +102,7 @@ account_t *account_new(const char *conffile, const char *id)
     a->add_missing_from_header = 1;
     a->add_missing_date_header = 1;
     a->remove_bcc_headers = 1;
+    a->source_ip = NULL;
     return a;
 }
 
@@ -177,6 +179,8 @@ account_t *account_copy(account_t *acc)
         a->tls_priorities =
             acc->tls_priorities ? xstrdup(acc->tls_priorities) : NULL;
         a->logfile = acc->logfile ? xstrdup(acc->logfile) : NULL;
+        a->logfile_time_format =
+            acc->logfile_time_format ? xstrdup(acc->logfile_time_format) : NULL;
         a->syslog = acc->syslog ? xstrdup(acc->syslog) : NULL;
         a->aliases = acc->aliases ? xstrdup(acc->aliases) : NULL;
         a->proxy_host = acc->proxy_host ? xstrdup(acc->proxy_host) : NULL;
@@ -184,6 +188,7 @@ account_t *account_copy(account_t *acc)
         a->add_missing_from_header = acc->add_missing_from_header;
         a->add_missing_date_header = acc->add_missing_date_header;
         a->remove_bcc_headers = acc->remove_bcc_headers;
+        a->source_ip = acc->source_ip ? xstrdup(acc->source_ip) : NULL;
     }
     return a;
 }
@@ -222,9 +227,11 @@ void account_free(void *a)
         free(p->dsn_return);
         free(p->dsn_notify);
         free(p->logfile);
+        free(p->logfile_time_format);
         free(p->syslog);
         free(p->aliases);
         free(p->proxy_host);
+        free(p->source_ip);
         free(p);
     }
 }
@@ -395,7 +402,8 @@ int check_auth_arg(char *arg)
             || strcmp(arg, "gssapi") == 0
             || strcmp(arg, "external") == 0
             || strcmp(arg, "login") == 0
-            || strcmp(arg, "ntlm") == 0)
+            || strcmp(arg, "ntlm") == 0
+            || strcmp(arg, "oauthbearer") == 0)
     {
         l = strlen(arg);
         for (i = 0; i < l; i++)
@@ -676,6 +684,12 @@ void override_account(account_t *acc1, account_t *acc2)
         free(acc1->logfile);
         acc1->logfile = acc2->logfile ? xstrdup(acc2->logfile) : NULL;
     }
+    if (acc2->mask & ACC_LOGFILE_TIME_FORMAT)
+    {
+        free(acc1->logfile_time_format);
+        acc1->logfile_time_format =
+            acc2->logfile_time_format ? xstrdup(acc2->logfile_time_format) : NULL;
+    }
     if (acc2->mask & ACC_SYSLOG)
     {
         free(acc1->syslog);
@@ -703,6 +717,11 @@ void override_account(account_t *acc1, account_t *acc2)
     {
         acc1->add_missing_date_header = acc2->add_missing_date_header;
     }
+    if (acc2->mask & ACC_SOURCE_IP)
+    {
+        free(acc1->source_ip);
+        acc1->source_ip = acc2->source_ip ? xstrdup(acc2->source_ip) : NULL;
+    }
     acc1->mask |= acc2->mask;
 }
 
@@ -740,28 +759,6 @@ int check_account(account_t *acc, int sendmail_mode, char **errstr)
         *errstr = xasprintf(_("tls_cert_file requires tls_key_file"));
         return CONF_ESYNTAX;
     }
-    if (acc->tls_nocertcheck && acc->tls_trust_file)
-    {
-        *errstr = xasprintf(
-                _("cannot use tls_trust_file with tls_certcheck turned off"));
-        return CONF_ESYNTAX;
-    }
-    if (acc->tls_nocertcheck
-            && (acc->tls_sha256_fingerprint
-                || acc->tls_sha1_fingerprint || acc->tls_md5_fingerprint))
-    {
-        *errstr = xasprintf(
-                _("cannot use tls_fingerprint with tls_certcheck turned off"));
-        return CONF_ESYNTAX;
-    }
-    if (acc->tls_trust_file
-            && (acc->tls_sha256_fingerprint
-                || acc->tls_sha1_fingerprint || acc->tls_md5_fingerprint))
-    {
-        *errstr = xasprintf(
-                _("cannot use both tls_trust_file and tls_fingerprint"));
-        return CONF_ESYNTAX;
-    }
     if (acc->tls && !acc->tls_trust_file
             && !acc->tls_sha256_fingerprint && !acc->tls_sha1_fingerprint
             && !acc->tls_md5_fingerprint && !acc->tls_nocertcheck)
@@ -791,6 +788,7 @@ int get_password_eval(const char *arg, char **buf, char **errstr)
 {
     FILE *eval;
     size_t l;
+    int have_more_data;
 
     *buf = NULL;
     *errstr = NULL;
@@ -802,33 +800,34 @@ int get_password_eval(const char *arg, char **buf, char **errstr)
         {
             errno = ENOMEM;
         }
-        *errstr = xasprintf("cannot evaluate '%s': %s", arg, strerror(errno));
+        *errstr = xasprintf(_("cannot evaluate '%s': %s"), arg, strerror(errno));
         return CONF_EIO;
     }
 
     *buf = xmalloc(LINEBUFSIZE);
     if (!fgets(*buf, LINEBUFSIZE, eval))
     {
-        *errstr = xasprintf("cannot read output of '%s'", arg);
+        *errstr = xasprintf(_("cannot read output of '%s'"), arg);
         pclose(eval);
         free(*buf);
         *buf = NULL;
         return CONF_EIO;
     }
+    have_more_data = (fgetc(eval) != EOF);
     pclose(eval);
 
     l = strlen(*buf);
     if (l > 0)
     {
-        if ((*buf)[l - 1] != '\n')
+        if ((*buf)[l - 1] != '\n' && have_more_data)
         {
-            *errstr = xasprintf("output of '%s' is longer than %d characters, "
-                    "or is not terminated by newline", arg, LINEBUFSIZE - 1);
+            *errstr = xasprintf(_("output of '%s' is longer than %d characters"),
+                    arg, LINEBUFSIZE - 1);
             free(*buf);
             *buf = NULL;
             return CONF_EIO;
         }
-        else
+        if ((*buf)[l - 1] == '\n')
         {
             (*buf)[l - 1] = '\0';
             if (l - 1 > 0 && (*buf)[l - 2] == '\r')
@@ -917,7 +916,7 @@ char *trim_string(const char *s)
  * whitespace) and an argument part (the word after the command).
  * Whitespace is ignored.
  * Sets the flag 'empty_line' if the line is empty.
- * Sets the flag 'eof' if EOF occured.
+ * Sets the flag 'eof' if EOF occurred.
  * On errors, 'empty_line' and 'eof', 'cmd' and 'arg' NULL.
  * On success, 'cmd' and 'arg' are allocated strings.
  * Used error codes: CONF_EIO, CONF_EPARSE
@@ -1139,6 +1138,7 @@ int read_conffile(const char *conffile, FILE *f, list_t **acc_list,
                 list_free(copy_from);
                 *errstr = xasprintf(_("line %d: missing account name"), line);
                 e = CONF_ESYNTAX;
+                free(acc_id);
                 break;
             }
             if (strchr(acc_id, ':') || strchr(acc_id, ','))
@@ -1147,15 +1147,17 @@ int read_conffile(const char *conffile, FILE *f, list_t **acc_list,
                 *errstr = xasprintf(_("line %d: an account name must not "
                             "contain colons or commas"), line);
                 e = CONF_ESYNTAX;
+                free(acc_id);
                 break;
             }
             if (find_account(*acc_list, acc_id))
             {
                 list_free(copy_from);
                 *errstr = xasprintf(
-                        _("line %d: account %s was already defined "
-                            "in this file"), line, arg);
+                        _("line %d: account %s was already defined"),
+                        line, acc_id);
                 e = CONF_ESYNTAX;
+                free(acc_id);
                 break;
             }
             acc = account_copy(defaults);
@@ -1624,6 +1626,19 @@ int read_conffile(const char *conffile, FILE *f, list_t **acc_list,
                 acc->logfile = expand_tilde(arg);
             }
         }
+        else if (strcmp(cmd, "logfile_time_format") == 0)
+        {
+            acc->mask |= ACC_LOGFILE_TIME_FORMAT;
+            free(acc->logfile_time_format);
+            if (*arg == '\0')
+            {
+                acc->logfile_time_format = NULL;
+            }
+            else
+            {
+                acc->logfile_time_format = xstrdup(arg);
+            }
+        }
         else if (strcmp(cmd, "syslog") == 0)
         {
             acc->mask |= ACC_SYSLOG;
@@ -1755,6 +1770,19 @@ int read_conffile(const char *conffile, FILE *f, list_t **acc_list,
                 break;
             }
         }
+        else if (strcmp(cmd, "source_ip") == 0)
+        {
+            acc->mask |= ACC_SOURCE_IP;
+            free(acc->source_ip);
+            if (*arg == '\0')
+            {
+                acc->source_ip = NULL;
+            }
+            else
+            {
+                acc->source_ip = xstrdup(arg);
+            }
+        }
         else if (strcmp(cmd, "keepbcc") == 0)
         {
             /* compatibility with 1.4.x */
