@@ -1,9 +1,10 @@
 /*
  * net.c
  *
- * This file is part of msmtp, an SMTP client.
+ * This file is part of msmtp, an SMTP client, and of mpop, a POP3 client.
  *
- * Copyright (C) 2000, 2003, 2004, 2005, 2006, 2007, 2008, 2012, 2014, 2015
+ * Copyright (C) 2000, 2003, 2004, 2005, 2006, 2007, 2008, 2012, 2014, 2015,
+ * 2018, 2019
  * Martin Lambers <marlam@marlam.de>
  *
  *   This program is free software; you can redistribute it and/or modify
@@ -25,9 +26,17 @@
 # include "config.h"
 #endif
 
+#ifdef W32_NATIVE
+# define WIN32_LEAN_AND_MEAN    /* do not include more than necessary */
+# define _WIN32_WINNT 0x0601    /* Windows 7 or later */
+# include <winsock2.h>
+# include <ws2tcpip.h>
+#endif
 #include <stdlib.h>
 #include <stdio.h>
+#include <stdint.h>
 #include <string.h>
+#include <limits.h>
 #include <unistd.h>
 #include <fcntl.h>
 #include <errno.h>
@@ -36,21 +45,24 @@
 #ifdef HAVE_SYS_SOCKET_H
 # include <sys/socket.h>
 #endif
+#ifdef HAVE_NETINET_IN_H
+# include <netinet/in.h>
+#endif
 #ifdef HAVE_ARPA_INET_H
 # include <arpa/inet.h>
 #endif
 #ifdef HAVE_NETDB_H
 # include <netdb.h>
 #endif
-#ifdef W32_NATIVE
-# define WIN32_LEAN_AND_MEAN    /* do not include more than necessary */
-# define _WIN32_WINNT 0x0502    /* Windows XP SP2 or later */
-# include <winsock2.h>
-# include <ws2tcpip.h>
-#endif
 
 #ifdef HAVE_LIBIDN
-# include <idna.h>
+# include <idn2.h>
+#endif
+
+#ifdef HAVE_LIBRESOLV
+# include <netinet/in.h>
+# include <arpa/nameser.h>
+# include <resolv.h>
 #endif
 
 #include "gettext.h"
@@ -58,6 +70,7 @@
 
 #include "xalloc.h"
 #include "readbuf.h"
+#include "tools.h"
 #include "net.h"
 
 
@@ -178,7 +191,7 @@ int net_recv(int fd, void *buf, size_t len, char **errstr)
     {
 #ifdef W32_NATIVE
         int e = WSAGetLastError();
-        if (e == WSAEWOULDBLOCK)
+        if (e == WSAETIMEDOUT)
         {
             *errstr = xasprintf(_("network read error: %s"),
                     _("the operation timed out"));
@@ -225,7 +238,7 @@ int net_send(int fd, const void *buf, size_t len, char **errstr)
     {
 #ifdef W32_NATIVE
         int e = WSAGetLastError();
-        if (e == WSAEWOULDBLOCK)
+        if (e == WSAETIMEDOUT)
         {
             *errstr = xasprintf(_("network write error: %s"),
                     _("the operation timed out"));
@@ -304,6 +317,45 @@ void net_close_socket(int fd)
 }
 
 
+/*
+ * net_bind_source_ip_to_socket()
+ *
+ * This function binds a source IP (in string representation, either IPv6 or IPv4)
+ * to a socket. It behaves like bind() in terms of return value and errno.
+ */
+
+int net_bind_source_ip_to_socket(int fd, const char *source_ip)
+{
+    struct sockaddr_in6 sa6;
+    struct sockaddr_in sa4;
+
+    memset(&sa6, 0, sizeof(sa6));
+    if (inet_pton(AF_INET6, source_ip, &sa6.sin6_addr) != 0)
+    {
+        sa6.sin6_family = AF_INET6;
+        return bind(fd, &sa6, sizeof(sa6));
+    }
+    else
+    {
+        memset(&sa4, 0, sizeof(sa4));
+        if (inet_pton(AF_INET, source_ip, &sa4.sin_addr) != 0)
+        {
+            sa4.sin_family = AF_INET;
+            return bind(fd, &sa4, sizeof(sa4));
+        }
+        else
+        {
+#ifdef W32_NATIVE
+            WSASetLastError(WSAEINVAL);
+#else
+            errno = EINVAL;
+#endif
+            return -1;
+        }
+    }
+}
+
+
 /*
  * net_connect()
  *
@@ -318,16 +370,17 @@ int net_connect(int fd, const struct sockaddr *serv_addr, socklen_t addrlen,
         int timeout)
 {
 #ifdef W32_NATIVE
-    /* TODO: I don't know how to do this on Win32. Please send a patch. */
-    return connect(fd, serv_addr, addrlen);
-#else /* UNIX or DJGPP */
-
+    u_long flags;
+    DWORD err;
+    int optlen;
+    fd_set eset;
+#else
     int flags;
-    struct timeval tv;
-    fd_set rset;
-    fd_set wset;
     int err;
     socklen_t optlen;
+#endif
+    struct timeval tv;
+    fd_set wset;
 
     if (timeout <= 0)
     {
@@ -336,8 +389,13 @@ int net_connect(int fd, const struct sockaddr *serv_addr, socklen_t addrlen,
     else
     {
         /* make socket non-blocking */
+#ifdef W32_NATIVE
+        flags = 1;
+        if (ioctlsocket(fd, FIONBIO, &flags) == SOCKET_ERROR)
+#else
         flags = fcntl(fd, F_GETFL, 0);
         if (fcntl(fd, F_SETFL, flags | O_NONBLOCK) == -1)
+#endif
         {
             return -1;
         }
@@ -345,51 +403,75 @@ int net_connect(int fd, const struct sockaddr *serv_addr, socklen_t addrlen,
         /* start connect */
         if (connect(fd, serv_addr, addrlen) < 0)
         {
+#ifdef W32_NATIVE
+            if (WSAGetLastError() != WSAEWOULDBLOCK)
+#else
             if (errno != EINPROGRESS)
+#endif
             {
                 return -1;
             }
 
             tv.tv_sec = timeout;
             tv.tv_usec = 0;
-            FD_ZERO(&rset);
             FD_ZERO(&wset);
-            FD_SET(fd, &rset);
             FD_SET(fd, &wset);
+#ifdef W32_NATIVE
+            FD_ZERO(&eset);
+            FD_SET(fd, &eset);
+#endif
 
             /* wait for connect() to finish */
-            if ((err = select(fd + 1, &rset, &wset, NULL, &tv)) <= 0)
+#ifdef W32_NATIVE
+            /* In case of an error on connect(), eset will be affected instead
+             * of wset (on Windows only). */
+            if ((err = select(fd + 1, NULL, &wset, &eset, &tv)) <= 0)
+#else
+            if ((err = select(fd + 1, NULL, &wset, NULL, &tv)) <= 0)
+#endif
             {
                 /* errno is already set if err < 0 */
                 if (err == 0)
                 {
+#ifdef W32_NATIVE
+                    WSASetLastError(WSAETIMEDOUT);
+#else
                     errno = ETIMEDOUT;
+#endif
                 }
                 return -1;
             }
 
             /* test for success, set errno */
-            optlen = sizeof(int);
+            optlen = sizeof(err);
             if (getsockopt(fd, SOL_SOCKET, SO_ERROR, &err, &optlen) < 0)
             {
                 return -1;
             }
             if (err != 0)
             {
+#ifdef W32_NATIVE
+                WSASetLastError(err);
+#else
                 errno = err;
+#endif
                 return -1;
             }
         }
 
         /* restore blocking mode */
+#ifdef W32_NATIVE
+        flags = 0;
+        if (ioctlsocket(fd, FIONBIO, &flags) == SOCKET_ERROR)
+#else
         if (fcntl(fd, F_SETFL, flags) == -1)
+#endif
         {
             return -1;
         }
 
         return 0;
     }
-#endif /* UNIX */
 }
 
 
@@ -410,7 +492,7 @@ void net_set_io_timeout(int socket, int seconds)
         (void)setsockopt(socket, SOL_SOCKET, SO_RCVTIMEO, &milliseconds, sizeof(int));
         (void)setsockopt(socket, SOL_SOCKET, SO_SNDTIMEO, &milliseconds, sizeof(int));
     }
-#else /* UNIX or DJGPP */
+#else /* UNIX */
     struct timeval tv;
 
     if (seconds > 0)
@@ -433,8 +515,7 @@ void net_set_io_timeout(int socket, int seconds)
 int net_socks5_connect(int fd, const char *hostname, int port, char **errstr)
 {
     /* maximum size of a SOCKS5 message (send or receive) */
-    const size_t required_buffer_size = 1 + 1 + 1 + 1 + 1 + 255 + 2;
-    unsigned char buffer[required_buffer_size];
+    unsigned char buffer[1 + 1 + 1 + 1 + 1 + 255 + 2];
     size_t hostname_len = strlen(hostname);
     uint16_t nport = htons(port);
     size_t len;
@@ -571,6 +652,7 @@ int net_socks5_connect(int fd, const char *hostname, int port, char **errstr)
 int net_open_socket(
         const char *proxy_hostname, int proxy_port,
         const char *hostname, int port,
+        const char *source_ip,
         int timeout,
         int *ret_fd, char **canonical_name, char **address,
         char **errstr)
@@ -589,7 +671,7 @@ int net_open_socket(
     if (proxy_hostname)
     {
         error_code = net_open_socket(NULL, -1, proxy_hostname, proxy_port,
-                timeout, &fd, NULL, NULL, errstr);
+                NULL, timeout, &fd, NULL, NULL, errstr);
         if (error_code != NET_EOK)
         {
             return error_code;
@@ -625,7 +707,7 @@ int net_open_socket(
     hints.ai_flags |= AI_IDN;
 # endif
 #elif defined(HAVE_LIBIDN)
-    idna_to_ascii_lz(hostname, &idn_hostname, 0);
+    idn2_to_ascii_lz(hostname, &idn_hostname, IDN2_NFC_INPUT | IDN2_NONTRANSITIONAL);
 #endif
     error_code = getaddrinfo(idn_hostname ? idn_hostname : hostname,
             port_string, &hints, &res0);
@@ -667,9 +749,21 @@ int net_open_socket(
 #endif
             continue;
         }
-        if (net_connect(fd, res->ai_addr, res->ai_addrlen, timeout) < 0)
+        if (source_ip && net_bind_source_ip_to_socket(fd, source_ip) != 0)
         {
             cause = 2;
+#ifdef W32_NATIVE
+            failure_errno = WSAGetLastError();
+#else
+            failure_errno = errno;
+#endif
+            net_close_socket(fd);
+            fd = -1;
+            continue;
+        }
+        if (net_connect(fd, res->ai_addr, res->ai_addrlen, timeout) < 0)
+        {
+            cause = 3;
 #ifdef W32_NATIVE
             if (WSAGetLastError() != WSAENETUNREACH)
             {
@@ -681,7 +775,7 @@ int net_open_socket(
                 failure_errno = errno;
             }
 #endif
-            close(fd);
+            net_close_socket(fd);
             fd = -1;
             continue;
         }
@@ -731,7 +825,18 @@ int net_open_socket(
                     );
             return NET_ESOCKET;
         }
-        else /* cause == 2 */
+        else if (cause == 2)
+        {
+            *errstr = xasprintf(_("cannot bind source ip %s: %s"), source_ip,
+#ifdef W32_NATIVE
+                    wsa_strerror(failure_errno)
+#else
+                    strerror(failure_errno)
+#endif
+                    );
+            return NET_ESOCKET;
+        }
+        else /* cause == 3 */
         {
 #ifdef W32_NATIVE
             if (failure_errno == 0)
@@ -906,6 +1011,89 @@ char *net_get_canonical_hostname(void)
 }
 
 
+/*
+ * net_get_srv_query()
+ *
+ * see net.h
+ */
+char* net_get_srv_query(const char *domain, const char *service)
+{
+    size_t domain_len = strlen(domain);
+    size_t service_len = strlen(service);
+    size_t query_len = 1 /* '_' */ + service_len + 6 /* "._tcp." */ + domain_len;
+    char* query = xmalloc(query_len + 1);
+    query[0] = '_';
+    strncpy(query + 1, service, service_len);
+    strncpy(query + 1 + service_len, "._tcp.", 6);
+    strcpy(query + 1 + service_len + 6, domain);
+    return query;
+}
+
+
+/*
+ * net_get_srv()
+ *
+ * see net.h
+ */
+int net_get_srv_record(const char* query, char **hostname, int *port)
+{
+#ifdef HAVE_LIBRESOLV
+
+    unsigned char buffer[NS_PACKETSZ];
+    int response_len;
+    ns_msg msg;
+    int i;
+    int current_prio = INT_MAX;
+    int current_weight = -1;
+    char *current_hostname = NULL;
+    int current_port = 0;
+
+    response_len = res_query(query, ns_c_in, ns_t_srv, buffer, sizeof(buffer));
+    if (response_len < 0) {
+        return NET_ESRVNOTFOUND;
+    }
+
+    ns_initparse(buffer, response_len, &msg);
+
+    for (i = 0; i < ns_msg_count(msg, ns_s_an); i++) {
+        ns_rr rr;
+        if (ns_parserr(&msg, ns_s_an, i, &rr))
+            continue; /* don't know what's wrong; ignore this part */
+        if (ns_rr_type(rr) == ns_t_srv) {
+            char name[NI_MAXHOST];
+            int prio, weight;
+            if (dn_expand(ns_msg_base(msg), ns_msg_end(msg), ns_rr_rdata(rr) + 6, name, sizeof(name)) < 0)
+                continue; /* don't know what's wrong; ignore this part */
+            if (name[0] == '\0')
+                continue; /* empty host name; ignore this part */
+            prio = ntohs(*((unsigned short*)ns_rr_rdata(rr) + 0));
+            weight = ntohs(*((unsigned short*)ns_rr_rdata(rr) + 1));
+            if (prio < current_prio || (prio == current_prio && weight > current_weight)) {
+                free(current_hostname);
+                current_hostname = xstrdup(name);
+                current_port = ntohs(*((unsigned short*)ns_rr_rdata(rr) + 2));
+                current_prio = prio;
+                current_weight = weight;
+            }
+        }
+    }
+    if (!current_hostname) {
+        /* the loop finished but we did not find usable information */
+        return NET_EIO;
+    } else {
+        *hostname = current_hostname;
+        *port = current_port;
+        return NET_EOK;
+    }
+
+#else
+
+    return NET_ELIBFAILED;
+
+#endif
+}
+
+
 /*
  * net_lib_deinit()
  *
@@ -918,3 +1106,30 @@ void net_lib_deinit(void)
     (void)WSACleanup();
 #endif
 }
+
+
+/*
+ * net_exitcode()
+ *
+ * see net.h
+ */
+
+int net_exitcode(int net_error_code)
+{
+    switch (net_error_code)
+    {
+        case NET_EHOSTNOTFOUND:
+            return EX_NOHOST;
+        case NET_ESOCKET:
+            return EX_OSERR;
+        case NET_ECONNECT:
+            return EX_TEMPFAIL;
+        case NET_EIO:
+            return EX_IOERR;
+        case NET_EPROXY:
+            return EX_UNAVAILABLE;
+        case NET_ELIBFAILED:
+        default:
+            return EX_SOFTWARE;
+    }
+}
