@@ -38,9 +38,9 @@
 #ifdef HAVE_LIBGSASL
 # include <gsasl.h>
 #else
-# include "base64.h"
 # include "md5-apps.h"
 #endif
+#include "base64.h"
 
 #include "gettext.h"
 #define _(string) gettext(string)
@@ -518,6 +518,10 @@ int smtp_init(smtp_server_t *srv, const char *ehlo_domain, list_t **errmsg,
             {
                 srv->cap.flags |= SMTP_CAP_AUTH_NTLM;
             }
+            if (strstr(s + 9, "OAUTHBEARER"))
+            {
+                srv->cap.flags |= SMTP_CAP_AUTH_OAUTHBEARER;
+            }
         }
         else if (strncmp(s + 4, "ETRN", 4) == 0)
         {
@@ -538,18 +542,21 @@ int smtp_init(smtp_server_t *srv, const char *ehlo_domain, list_t **errmsg,
 
 #ifdef HAVE_TLS
 int smtp_tls_init(smtp_server_t *srv,
-        const char *tls_key_file, const char *tls_cert_file,
+        const char *tls_key_file, const char *tls_cert_file, const char *pin,
         const char *tls_trust_file, const char *tls_crl_file,
         const unsigned char *tls_sha256_fingerprint,
         const unsigned char *tls_sha1_fingerprint,
         const unsigned char *tls_md5_fingerprint,
         int min_dh_prime_bits,
-        const char *priorities, char **errstr)
+        const char *priorities,
+        const char *hostname,
+        int no_certcheck,
+        char **errstr)
 {
-    return tls_init(&srv->tls, tls_key_file, tls_cert_file,
+    return tls_init(&srv->tls, tls_key_file, tls_cert_file, pin,
             tls_trust_file, tls_crl_file,
             tls_sha256_fingerprint, tls_sha1_fingerprint, tls_md5_fingerprint,
-            min_dh_prime_bits, priorities, errstr);
+            min_dh_prime_bits, priorities, hostname, no_certcheck, errstr);
 }
 #endif /* HAVE_TLS */
 
@@ -599,10 +606,10 @@ int smtp_tls_starttls(smtp_server_t *srv, list_t **error_msg, char **errstr)
  */
 
 #ifdef HAVE_TLS
-int smtp_tls(smtp_server_t *srv, const char *hostname, int tls_nocertcheck,
+int smtp_tls(smtp_server_t *srv,
         tls_cert_info_t *tci, char **tls_parameter_description, char **errstr)
 {
-    return tls_start(&srv->tls, srv->fd, hostname, tls_nocertcheck, tci,
+    return tls_start(&srv->tls, srv->fd, tci,
             tls_parameter_description, errstr);
 }
 #endif /* HAVE_TLS */
@@ -916,6 +923,79 @@ int smtp_auth_external(smtp_server_t *srv, const char *user,
 #endif /* !HAVE_LIBGSASL */
 
 
+/*
+ * smtp_auth_oauthbearer()
+ *
+ * Do SMTP authentication via AUTH OAUTHBEARER.
+ * The SMTP server must support SMTP_CAP_AUTH_OAUTHBEARER
+ * Used error codes: SMTP_EIO, SMTP_EAUTHFAIL, SMTP_EINVAL
+ */
+
+int smtp_auth_oauthbearer(smtp_server_t *srv,
+        const char *hostname, unsigned short port,
+        const char *user, const char *token,
+        list_t **error_msg, char **errstr)
+{
+    list_t *msg;
+    char *oauth;
+    size_t oa_len;
+    char *b64;
+    size_t b64_len;
+    int e;
+    int status;
+
+    *error_msg = NULL;
+
+    oa_len = 4 + /* "n,a=" */
+             strlen(user) +
+             7 + /* ",^Ahost=" */
+             strlen(hostname) +
+             6 + 5 + /* "^Aport=" + up to 5 digits */
+             13 + /* "^Aauth=Bearer " */
+             strlen(token) +
+             2 /* "^A^A" */;
+    oauth = xmalloc(oa_len + 1);
+    oa_len = snprintf(oauth, oa_len + 1,
+                      "n,a=%s,\001host=%s\001port=%d\001auth=Bearer %s\001\001",
+                      user, hostname, port, token);
+    b64_len = BASE64_LENGTH(oa_len) + 1;
+    b64 = xmalloc(b64_len);
+    base64_encode(oauth, oa_len, b64, b64_len);
+    e = smtp_send_cmd(srv, errstr, "AUTH OAUTHBEARER %s", b64);
+    free(oauth);
+    free(b64);
+    if (e != SMTP_EOK)
+    {
+        return e;
+    }
+    if ((e = smtp_get_msg(srv, &msg, errstr)) != SMTP_EOK)
+    {
+        return e;
+    }
+    if ((status = smtp_msg_status(msg)) != 235)
+    {
+        if (status == 334)
+        {
+            list_xfree(msg, free);
+            if ((e = smtp_send_cmd(srv, errstr, "")) != SMTP_EOK)
+            {
+                return e;
+            }
+            if ((e = smtp_get_msg(srv, &msg, errstr)) != SMTP_EOK)
+            {
+                return e;
+            }
+        }
+        *error_msg = msg;
+        *errstr = xasprintf(_("authentication failed (method %s)"), "OAUTHBEARER");
+        return SMTP_EAUTHFAIL;
+    }
+    list_xfree(msg, free);
+
+    return SMTP_EOK;
+}
+
+
 /*
  * smtp_server_supports_authmech()
  *
@@ -939,7 +1019,9 @@ int smtp_server_supports_authmech(smtp_server_t *srv, const char *mech)
             || ((srv->cap.flags & SMTP_CAP_AUTH_LOGIN)
                 && strcmp(mech, "LOGIN") == 0)
             || ((srv->cap.flags & SMTP_CAP_AUTH_NTLM)
-                && strcmp(mech, "NTLM") == 0));
+                && strcmp(mech, "NTLM") == 0)
+            || ((srv->cap.flags & SMTP_CAP_AUTH_OAUTHBEARER)
+                && strcmp(mech, "OAUTHBEARER") == 0));
 }
 
 
@@ -956,12 +1038,19 @@ int smtp_client_supports_authmech(const char *mech)
     int supported = 0;
     Gsasl *ctx;
 
-    if (gsasl_init(&ctx) != GSASL_OK)
+    if (strcmp(mech, "OAUTHBEARER") == 0)
     {
-        return 0;
+        supported = 1;
+    }
+    else
+    {
+        if (gsasl_init(&ctx) != GSASL_OK)
+        {
+            return 0;
+        }
+        supported = gsasl_client_support_p(ctx, mech);
+        gsasl_done(ctx);
     }
-    supported = gsasl_client_support_p(ctx, mech);
-    gsasl_done(ctx);
     return supported;
 
 #else /* not HAVE_LIBGSASL */
@@ -969,7 +1058,8 @@ int smtp_client_supports_authmech(const char *mech)
     return (strcmp(mech, "CRAM-MD5") == 0
             || strcmp(mech, "PLAIN") == 0
             || strcmp(mech, "EXTERNAL") == 0
-            || strcmp(mech, "LOGIN") == 0);
+            || strcmp(mech, "LOGIN") == 0
+            || strcmp(mech, "OAUTHBEARER") == 0);
 
 #endif /* not HAVE_LIBGSASL */
 }
@@ -983,6 +1073,7 @@ int smtp_client_supports_authmech(const char *mech)
 
 int smtp_auth(smtp_server_t *srv,
         const char *hostname,
+        unsigned short port,
         const char *user,
         const char *password,
         const char *ntlmdomain,
@@ -1018,7 +1109,7 @@ int smtp_auth(smtp_server_t *srv,
         *errstr = xasprintf(_("GNU SASL: %s"), gsasl_strerror(error_code));
         return SMTP_ELIBFAILED;
     }
-    if (strcmp(auth_mech, "") != 0 && !gsasl_client_support_p(ctx, auth_mech))
+    if (strcmp(auth_mech, "") != 0 && !smtp_client_supports_authmech(auth_mech))
     {
         gsasl_done(ctx);
         *errstr = xasprintf(
@@ -1096,8 +1187,7 @@ int smtp_auth(smtp_server_t *srv,
     /* Check availability of required authentication data */
     if (strcmp(auth_mech, "EXTERNAL") != 0)
     {
-        /* GSSAPI, SCRAM-SHA-1, DIGEST-MD5, CRAM-MD5, PLAIN, LOGIN, NTLM all
-         * need a user name */
+        /* All authentication schemes need a user name */
         if (!user)
         {
             gsasl_done(ctx);
@@ -1105,8 +1195,8 @@ int smtp_auth(smtp_server_t *srv,
                     auth_mech);
             return SMTP_EUNAVAIL;
         }
-        /* SCRAM-SHA-1, DIGEST-MD5, CRAM-MD5, PLAIN, LOGIN, NTLM all need a
-         * password */
+        /* SCRAM-SHA-1, DIGEST-MD5, CRAM-MD5, PLAIN, LOGIN, NTLM and OAUTHBEARER
+         * all need a password */
         if (strcmp(auth_mech, "GSSAPI") != 0 && !password)
         {
             if (!password_callback
@@ -1122,7 +1212,16 @@ int smtp_auth(smtp_server_t *srv,
         }
     }
 
-    if ((error_code = gsasl_client_start(ctx, auth_mech, &sctx)) != GSASL_OK)
+    /* OAUTHBEARER is built-in, all other methods are provided by GNU SASL */
+    if (strcmp(auth_mech, "OAUTHBEARER") == 0)
+    {
+        gsasl_done(ctx);
+        e = smtp_auth_oauthbearer(srv, hostname, port, user, password,
+                                  error_msg, errstr);
+        free(callback_password);
+        return e;
+    }
+    else if ((error_code = gsasl_client_start(ctx, auth_mech, &sctx)) != GSASL_OK)
     {
         gsasl_done(ctx);
         *errstr = xasprintf(_("GNU SASL: %s"), gsasl_strerror(error_code));
@@ -1359,7 +1458,8 @@ int smtp_auth(smtp_server_t *srv,
 
     if (strcmp(auth_mech, "EXTERNAL") != 0)
     {
-        /* CRAMD-MD5, PLAIN, LOGIN all need a user name and a password */
+        /* CRAMD-MD5, PLAIN, LOGIN, OAUTHBEARER all need a user name and a
+         * password */
         if (!user)
         {
             *errstr = xasprintf(_("authentication method %s needs a user name"),
@@ -1396,6 +1496,11 @@ int smtp_auth(smtp_server_t *srv,
     {
         e = smtp_auth_login(srv, user, password, error_msg, errstr);
     }
+    else if (strcmp(auth_mech, "OAUTHBEARER") == 0)
+    {
+        e = smtp_auth_oauthbearer(srv, hostname, port, user, password,
+                                  error_msg, errstr);
+    }
     else
     {
         *errstr = xasprintf(_("authentication method %s not supported"),
