@@ -3,7 +3,8 @@
  *
  * This file is part of msmtp, an SMTP client.
  *
- * Copyright (C) 2000, 2003, 2004, 2005, 2006, 2008, 2010, 2012, 2014, 2016
+ * Copyright (C) 2000, 2003, 2004, 2005, 2006, 2008, 2010, 2012, 2014, 2016,
+ * 2018, 2019
  * Martin Lambers <marlam@marlam.de>
  *
  *   This program is free software; you can redistribute it and/or modify
@@ -71,7 +72,8 @@
 #define SMTP_CAP_AUTH_GSSAPI            (1 << 10)
 #define SMTP_CAP_AUTH_EXTERNAL          (1 << 11)
 #define SMTP_CAP_AUTH_NTLM              (1 << 12)
-#define SMTP_CAP_ETRN                   (1 << 13)
+#define SMTP_CAP_AUTH_OAUTHBEARER       (1 << 13)
+#define SMTP_CAP_ETRN                   (1 << 14)
 
 
 /*
@@ -127,7 +129,7 @@ smtp_server_t smtp_new(FILE *debug, int protocol);
  * Success: NET_EOK
  */
 int smtp_connect(smtp_server_t *srv, const char *proxy_host, int proxy_port,
-        const char *host, int port, int timeout,
+        const char *host, int port, const char *source_ip, int timeout,
         char **server_canonical_name, char **server_address,
         char **errstr);
 
@@ -175,13 +177,16 @@ int smtp_init(smtp_server_t *srv, const char *ehlo_domain, list_t **msg,
  */
 #ifdef HAVE_TLS
 int smtp_tls_init(smtp_server_t *srv,
-        const char *tls_key_file, const char *tls_cert_file,
+        const char *tls_key_file, const char *tls_cert_file, const char *pin,
         const char *tls_trust_file, const char *tls_crl_file,
         const unsigned char *tls_sha256_fingerprint,
         const unsigned char *tls_sha1_fingerprint,
         const unsigned char *tls_md5_fingerprint,
         int min_dh_prime_bits,
-        const char *priorities, char **errstr);
+        const char *priorities,
+        const char *hostname,
+        int no_certcheck,
+        char **errstr);
 #endif /* HAVE_TLS */
 
 /*
@@ -214,8 +219,8 @@ int smtp_tls_starttls(smtp_server_t *srv, list_t **error_msg, char **errstr);
  * Success: TLS_EOK
  */
 #ifdef HAVE_TLS
-int smtp_tls(smtp_server_t *srv, const char *hostname, int tls_nocertcheck,
-        tls_cert_info_t *tci, char **errstr);
+int smtp_tls(smtp_server_t *srv,
+        tls_cert_info_t *tci, char **tls_parameter_description, char **errstr);
 #endif /* HAVE_TLS */
 
 /*
@@ -246,6 +251,8 @@ int smtp_server_supports_authmech(smtp_server_t *srv, const char *mech);
  * SMTP_EINSECURE.
  * The hostname is the name of the SMTP server. It may be needed for
  * authentication.
+ * The port is port number SMTP server accepts connections on. It may be needed
+ * for authentication.
  * The ntlmdomain may be NULL (even if you use NTLM authentication).
  * If 'password' is NULL, but the authentication method needs a password,
  * the 'password_callback' function is called (if 'password_callback' is not
@@ -257,6 +264,7 @@ int smtp_server_supports_authmech(smtp_server_t *srv, const char *mech);
  */
 int smtp_auth(smtp_server_t *srv,
         const char *hostname,
+        unsigned short port,
         const char *user,
         const char *password,
         const char *ntlmdomain,
@@ -340,7 +348,7 @@ int smtp_end_mail(smtp_server_t *srv, list_t **msg, char **errstr);
  * the same list that was given to smtp_send_envelope()).
  *
  * If all of these replies are positive, SMTP_EOK will be returned.
- * If an IO error occured, SMTP_EIO will be returned (as always).
+ * If an IO error occurred, SMTP_EIO will be returned (as always).
  * In both cases, 'errstrs' and 'error_msgs' will be NULL.
  *
  * If one or more of the replies are negative, SMTP_EUNAVAIL will be returned,
@@ -386,4 +394,11 @@ int smtp_quit(smtp_server_t *srv, char **errstr);
  */
 void smtp_close(smtp_server_t *srv);
 
+/*
+ * smtp_exitcode()
+ *
+ * Translate SMTP_* error code to an error code from sysexits.h
+ */
+int smtp_exitcode(int smtp_error_code);
+
 #endif
