@@ -4,7 +4,7 @@
  * This file is part of msmtp, an SMTP client.
  *
  * Copyright (C) 2000, 2003, 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011,
- * 2012, 2014, 2016
+ * 2012, 2014, 2016, 2018, 2019
  * Martin Lambers <marlam@marlam.de>
  *
  *   This program is free software; you can redistribute it and/or modify
@@ -37,6 +37,7 @@
 #ifdef HAVE_LIBGNUTLS
 # include <gnutls/gnutls.h>
 # include <gnutls/x509.h>
+# include <gnutls/pkcs11.h>
 #endif /* HAVE_LIBGNUTLS */
 #ifdef HAVE_LIBSSL
 # include <openssl/ssl.h>
@@ -45,17 +46,23 @@
 # include <openssl/err.h>
 # include <openssl/rand.h>
 # include <openssl/evp.h>
+#if OPENSSL_VERSION_NUMBER < 0x10100000L
+#define X509_get0_notBefore X509_get_notBefore
+#define X509_get0_notAfter X509_get_notAfter
+#endif
 #endif /* HAVE_LIBSSL */
 
 #ifdef HAVE_LIBIDN
-# include <idna.h>
+# include <idn2.h>
 #endif
 
 #include "gettext.h"
 #define _(string) gettext(string)
+#define N_(string) gettext_noop(string)
 
 #include "xalloc.h"
 #include "readbuf.h"
+#include "tools.h"
 #include "tls.h"
 
 
@@ -72,6 +79,8 @@ void tls_clear(tls_t *tls)
     tls->have_sha256_fingerprint = 0;
     tls->have_sha1_fingerprint = 0;
     tls->have_md5_fingerprint = 0;
+    tls->no_certcheck = 0;
+    tls->hostname = NULL;
 }
 
 
@@ -109,12 +118,7 @@ int seed_prng(char **errstr)
         }
         /* Seed in time. I can't think of other "random" things on DOS
          * systems. */
-        if ((t = time(NULL)) < 0)
-        {
-            *errstr = xasprintf(_("cannot get system time: %s"),
-                    strerror(errno));
-            return TLS_ESEED;
-        }
+        t = time(NULL);
         RAND_seed((unsigned char *)&t, sizeof(time_t));
         /* If the RANDFILE + time is not enough, we fall back to the insecure
          * and stupid method of seeding OpenSSLs PRNG with the systems PRNG. */
@@ -167,8 +171,10 @@ int tls_lib_init(char **errstr)
 #ifdef HAVE_LIBSSL
     int e;
 
+#if OPENSSL_VERSION_NUMBER < 0x10100000L
     SSL_load_error_strings();
     SSL_library_init();
+#endif
     if ((e = seed_prng(errstr)) != TLS_EOK)
     {
         return e;
@@ -473,7 +479,7 @@ int tls_cert_info_get(tls_t *tls, tls_cert_info_t *tci, char **errstr)
     X509 *x509cert;
     X509_NAME *x509_subject;
     X509_NAME *x509_issuer;
-    ASN1_TIME *asn1time;
+    const ASN1_TIME *asn1time;
     int nid[6] = { NID_commonName,
         NID_organizationName,
         NID_organizationalUnitName,
@@ -518,8 +524,8 @@ int tls_cert_info_get(tls_t *tls, tls_cert_info_t *tci, char **errstr)
         *errstr = xasprintf(_("%s: error getting SHA1 fingerprint"), errmsg);
         return TLS_ECERT;
     }
-    asn1time = X509_get_notBefore(x509cert);
-    if (asn1time_to_time_t((char *)asn1time->data,
+    asn1time = X509_get0_notBefore(x509cert);
+    if (asn1time_to_time_t((const char *)asn1time->data,
                 (asn1time->type != V_ASN1_GENERALIZEDTIME),
                 &(tci->activation_time)) != 0)
     {
@@ -528,8 +534,8 @@ int tls_cert_info_get(tls_t *tls, tls_cert_info_t *tci, char **errstr)
         tls_cert_info_free(tci);
         return TLS_ECERT;
     }
-    asn1time = X509_get_notAfter(x509cert);
-    if (asn1time_to_time_t((char *)asn1time->data,
+    asn1time = X509_get0_notAfter(x509cert);
+    if (asn1time_to_time_t((const char *)asn1time->data,
                 (asn1time->type != V_ASN1_GENERALIZEDTIME),
                 &(tci->expiration_time)) != 0)
     {
@@ -577,6 +583,81 @@ int tls_cert_info_get(tls_t *tls, tls_cert_info_t *tci, char **errstr)
 }
 
 
+/*
+ * tls_print_info()
+ *
+ * see tls.h
+ */
+
+/* Convert the given time into a string. */
+static void tls_time_to_string(const time_t *t, char *buf, size_t bufsize)
+{
+#ifdef ENABLE_NLS
+    (void)strftime(buf, bufsize, "%c", localtime(t));
+#else
+    char *p;
+
+    (void)snprintf(buf, bufsize, "%s", ctime(t));
+    if ((p = strchr(buf, '\n')))
+    {
+        *p = '\0';
+    }
+#endif
+}
+
+void tls_print_info(const char *tls_parameter_description,
+        const tls_cert_info_t *tci)
+{
+    const char *info_fieldname[6] = { N_("Common Name"), N_("Organization"),
+        N_("Organizational unit"), N_("Locality"), N_("State or Province"),
+        N_("Country") };
+    char sha256_fingerprint_string[96];
+    char sha1_fingerprint_string[60];
+    char timebuf[128];          /* should be long enough for every locale */
+    char *tmp;
+    int i;
+
+    printf(_("TLS session parameters:\n"));
+    printf("    %s\n", tls_parameter_description
+            ? tls_parameter_description : _("not available"));
+
+    print_fingerprint(sha256_fingerprint_string, tci->sha256_fingerprint, 32);
+    print_fingerprint(sha1_fingerprint_string, tci->sha1_fingerprint, 20);
+
+    printf(_("TLS certificate information:\n"));
+    printf("    %s:\n", _("Owner"));
+    for (i = 0; i < 6; i++)
+    {
+        if (tci->owner_info[i])
+        {
+            tmp = xstrdup(tci->owner_info[i]);
+            printf("        %s: %s\n", gettext(info_fieldname[i]),
+                    sanitize_string(tmp));
+            free(tmp);
+        }
+    }
+    printf("    %s:\n", _("Issuer"));
+    for (i = 0; i < 6; i++)
+    {
+        if (tci->issuer_info[i])
+        {
+            tmp = xstrdup(tci->issuer_info[i]);
+            printf("        %s: %s\n", gettext(info_fieldname[i]),
+                    sanitize_string(tmp));
+            free(tmp);
+        }
+    }
+    printf("    %s:\n", _("Validity"));
+    tls_time_to_string(&tci->activation_time, timebuf, sizeof(timebuf));
+    printf("        %s: %s\n", _("Activation time"), timebuf);
+    tls_time_to_string(&tci->expiration_time, timebuf, sizeof(timebuf));
+    printf("        %s: %s\n", _("Expiration time"), timebuf);
+    printf("    %s:\n", _("Fingerprints"));
+    printf("        SHA256: %s\n", sha256_fingerprint_string);
+    printf("        SHA1 (deprecated): %s\n", sha1_fingerprint_string);
+}
+
+
 /*
  * [OpenSSL only] hostname_match()
  *
@@ -651,7 +732,7 @@ int hostname_match(const char *hostname, const char *certname)
  * Used error codes: TLS_ECERT
  */
 
-int tls_check_cert(tls_t *tls, const char *hostname, char **errstr)
+int tls_check_cert(tls_t *tls, char **errstr)
 {
 #ifdef HAVE_LIBGNUTLS
     int error_code;
@@ -767,16 +848,21 @@ int tls_check_cert(tls_t *tls, const char *hostname, char **errstr)
     /* If 'tls->have_trust_file' is true, this function uses the trusted CAs
      * in the credentials structure. So you must have installed one or more CA
      * certificates. */
-    gnutls_certificate_set_verify_flags(tls->cred,
-            GNUTLS_VERIFY_ALLOW_X509_V1_CA_CRT);
     if ((error_code = gnutls_certificate_verify_peers2(tls->session,
                     &status)) != 0)
     {
         *errstr = xasprintf("%s: %s", error_msg, gnutls_strerror(error_code));
         return TLS_ECERT;
     }
+    if (gnutls_certificate_type_get(tls->session) != GNUTLS_CRT_X509)
+    {
+        *errstr = xasprintf(_("%s: the certificate type is not X509"),
+                error_msg);
+        return TLS_ECERT;
+    }
     if (tls->have_trust_file)
     {
+#if GNUTLS_VERSION_NUMBER < 0x030200
         if (status & GNUTLS_CERT_REVOKED)
         {
             *errstr = xasprintf(_("%s: the certificate has been revoked"),
@@ -796,12 +882,17 @@ int tls_check_cert(tls_t *tls, const char *hostname, char **errstr)
                     error_msg);
             return TLS_ECERT;
         }
-    }
-    if (gnutls_certificate_type_get(tls->session) != GNUTLS_CRT_X509)
-    {
-        *errstr = xasprintf(_("%s: the certificate type is not X509"),
-                error_msg);
-        return TLS_ECERT;
+#else
+        if (status)
+        {
+            gnutls_datum_t txt;
+            gnutls_certificate_verification_status_print(status,
+                    GNUTLS_CRT_X509, &txt, 0);
+            *errstr = xasprintf(_("%s: %s"), error_msg, txt.data);
+            gnutls_free(txt.data);
+            return TLS_ECERT;
+        }
+#endif
     }
     if (!(cert_list = gnutls_certificate_get_peers(
                     tls->session, &cert_list_size)))
@@ -810,12 +901,7 @@ int tls_check_cert(tls_t *tls, const char *hostname, char **errstr)
         return TLS_ECERT;
     }
     /* Needed to check times: */
-    if ((t1 = time(NULL)) < 0)
-    {
-        *errstr = xasprintf("%s: cannot get system time: %s", error_msg,
-                strerror(errno));
-        return TLS_ECERT;
-    }
+    t1 = time(NULL);
     /* Check the certificate chain. All certificates in the chain must have
      * valid activation/expiration times. The first certificate in the chain is
      * the host's certificate; it must match the hostname. */
@@ -839,15 +925,15 @@ int tls_check_cert(tls_t *tls, const char *hostname, char **errstr)
         if (i == 0)
         {
 #if GNUTLS_VERSION_NUMBER < 0x030400 && defined(HAVE_LIBIDN)
-            idna_to_ascii_lz(hostname, &idn_hostname, 0);
+            idn2_to_ascii_lz(tls->hostname, &idn_hostname, IDN2_NFC_INPUT | IDN2_NONTRANSITIONAL);
 #endif
             error_code = gnutls_x509_crt_check_hostname(cert,
-                    idn_hostname ? idn_hostname : hostname);
+                    idn_hostname ? idn_hostname : tls->hostname);
             free(idn_hostname);
             if (error_code == 0)
             {
                 *errstr = xasprintf(_("%s: the certificate owner does not "
-                            "match hostname %s"), error_msg, hostname);
+                            "match hostname %s"), error_msg, tls->hostname);
                 return TLS_ECERT;
             }
         }
@@ -1005,7 +1091,7 @@ int tls_check_cert(tls_t *tls, const char *hostname, char **errstr)
      * type DNS or the Common Name (CN). */
 
 #ifdef HAVE_LIBIDN
-    idna_to_ascii_lz(hostname, &idn_hostname, 0);
+    idn2_to_ascii_lz(tls->hostname, &idn_hostname, IDN2_NFC_INPUT | IDN2_NONTRANSITIONAL);
 #endif
 
     /* Try the DNS subjectAltNames. */
@@ -1029,7 +1115,7 @@ int tls_check_cert(tls_t *tls, const char *hostname, char **errstr)
                     return TLS_ECERT;
                 }
                 if ((match_found = hostname_match(
-                                idn_hostname ? idn_hostname : hostname,
+                                idn_hostname ? idn_hostname : tls->hostname,
                                 (char *)(subj_alt_name->d.ia5->data))))
                 {
                     break;
@@ -1070,7 +1156,7 @@ int tls_check_cert(tls_t *tls, const char *hostname, char **errstr)
             free(buf);
             return TLS_ECERT;
         }
-        match_found = hostname_match(idn_hostname ? idn_hostname : hostname,
+        match_found = hostname_match(idn_hostname ? idn_hostname : tls->hostname,
                 buf);
         free(buf);
     }
@@ -1081,7 +1167,7 @@ int tls_check_cert(tls_t *tls, const char *hostname, char **errstr)
     {
         *errstr = xasprintf(
                 _("%s: the certificate owner does not match hostname %s"),
-                error_msg, hostname);
+                error_msg, tls->hostname);
         return TLS_ECERT;
     }
 
@@ -1090,6 +1176,34 @@ int tls_check_cert(tls_t *tls, const char *hostname, char **errstr)
 }
 
 
+/*
+ * tls_pin_callback()
+ *
+ * Passes a PIN to GnuTLS for PKCS11 smart cards or similar
+ */
+#ifdef HAVE_LIBGNUTLS
+static int tls_pin_callback(void *userdata, int attempt,
+        const char *token_url, const char *token_label,
+        unsigned int flags, char *pin, size_t pin_max)
+{
+    (void)attempt;
+    (void)token_url;
+    (void)token_label;
+    (void)flags;
+
+    size_t len;
+    if (userdata && (len = strlen(userdata)) < pin_max)
+    {
+        strcpy(pin, userdata);
+        return 0;
+    }
+    else
+    {
+        return 1;
+    }
+}
+#endif
+
 /*
  * tls_init()
  *
@@ -1097,12 +1211,14 @@ int tls_check_cert(tls_t *tls, const char *hostname, char **errstr)
  */
 
 int tls_init(tls_t *tls,
-        const char *key_file, const char *cert_file,
+        const char *key_file, const char *cert_file, const char *pin,
         const char *trust_file, const char *crl_file,
         const unsigned char *sha256_fingerprint,
         const unsigned char *sha1_fingerprint,
         const unsigned char *md5_fingerprint,
         int min_dh_prime_bits, const char *priorities,
+        const char *hostname,
+        int no_certcheck,
         char **errstr)
 {
 #ifdef HAVE_LIBGNUTLS
@@ -1164,6 +1280,7 @@ int tls_init(tls_t *tls,
     }
     if (key_file && cert_file)
     {
+        gnutls_pkcs11_set_pin_function(tls_pin_callback, (void*)pin);
         if ((error_code = gnutls_certificate_set_x509_key_file(tls->cred,
                         cert_file, key_file, GNUTLS_X509_FMT_PEM)) < 0)
         {
@@ -1175,44 +1292,64 @@ int tls_init(tls_t *tls,
             return TLS_EFILE;
         }
     }
-    if (trust_file)
+    if (trust_file
+            && !no_certcheck
+            && !sha256_fingerprint
+            && !sha1_fingerprint
+            && !md5_fingerprint)
     {
-        if ((error_code = gnutls_certificate_set_x509_trust_file(
-                        tls->cred, trust_file, GNUTLS_X509_FMT_PEM)) <= 0)
+        if (strcmp(trust_file, "system") == 0)
         {
-            *errstr = xasprintf(
-                    _("cannot set X509 trust file %s for TLS session: %s"),
-                    trust_file, gnutls_strerror(error_code));
-            gnutls_deinit(tls->session);
-            gnutls_certificate_free_credentials(tls->cred);
-            return TLS_EFILE;
+            if ((error_code = gnutls_certificate_set_x509_system_trust(
+                            tls->cred)) < 0)
+            {
+                *errstr = xasprintf(
+                        _("cannot set X509 system trust for TLS session: %s"),
+                        gnutls_strerror(error_code));
+                gnutls_deinit(tls->session);
+                gnutls_certificate_free_credentials(tls->cred);
+                return TLS_ELIBFAILED;
+            }
+        }
+        else
+        {
+            if ((error_code = gnutls_certificate_set_x509_trust_file(
+                            tls->cred, trust_file, GNUTLS_X509_FMT_PEM)) <= 0)
+            {
+                *errstr = xasprintf(
+                        _("cannot set X509 trust file %s for TLS session: %s"),
+                        trust_file, gnutls_strerror(error_code));
+                gnutls_deinit(tls->session);
+                gnutls_certificate_free_credentials(tls->cred);
+                return TLS_EFILE;
+            }
+        }
+        if (crl_file)
+        {
+            if ((error_code = gnutls_certificate_set_x509_crl_file(
+                            tls->cred, crl_file, GNUTLS_X509_FMT_PEM)) < 0)
+            {
+                *errstr = xasprintf(
+                        _("cannot set X509 CRL file %s for TLS session: %s"),
+                        crl_file, gnutls_strerror(error_code));
+                gnutls_deinit(tls->session);
+                gnutls_certificate_free_credentials(tls->cred);
+                return TLS_EFILE;
+            }
         }
         tls->have_trust_file = 1;
     }
-    if (trust_file && crl_file)
-    {
-        if ((error_code = gnutls_certificate_set_x509_crl_file(
-                        tls->cred, crl_file, GNUTLS_X509_FMT_PEM)) < 0)
-        {
-            *errstr = xasprintf(
-                    _("cannot set X509 CRL file %s for TLS session: %s"),
-                    crl_file, gnutls_strerror(error_code));
-            gnutls_deinit(tls->session);
-            gnutls_certificate_free_credentials(tls->cred);
-            return TLS_EFILE;
-        }
-    }
-    if (sha256_fingerprint)
+    if (sha256_fingerprint && !no_certcheck)
     {
         memcpy(tls->fingerprint, sha256_fingerprint, 32);
         tls->have_sha256_fingerprint = 1;
     }
-    else if (sha1_fingerprint)
+    else if (sha1_fingerprint && !no_certcheck)
     {
         memcpy(tls->fingerprint, sha1_fingerprint, 20);
         tls->have_sha1_fingerprint = 1;
     }
-    else if (md5_fingerprint)
+    else if (md5_fingerprint && !no_certcheck)
     {
         memcpy(tls->fingerprint, md5_fingerprint, 16);
         tls->have_md5_fingerprint = 1;
@@ -1226,6 +1363,8 @@ int tls_init(tls_t *tls,
         gnutls_certificate_free_credentials(tls->cred);
         return TLS_ELIBFAILED;
     }
+    tls->no_certcheck = no_certcheck;
+    tls->hostname = xstrdup(hostname);
     return TLS_EOK;
 
 #endif /* HAVE_LIBGNUTLS */
@@ -1292,29 +1431,47 @@ int tls_init(tls_t *tls,
             return TLS_EFILE;
         }
     }
-    if (trust_file)
+    if (trust_file
+            && !no_certcheck
+            && !sha256_fingerprint
+            && !sha1_fingerprint
+            && !md5_fingerprint)
     {
-        if (SSL_CTX_load_verify_locations(tls->ssl_ctx, trust_file, NULL) != 1)
+        if (strcmp(trust_file, "system") == 0)
         {
-            *errstr = xasprintf(_("cannot load trust file %s: %s"),
-                    trust_file, ERR_error_string(ERR_get_error(), NULL));
-            SSL_CTX_free(tls->ssl_ctx);
-            tls->ssl_ctx = NULL;
-            return TLS_EFILE;
+            if (SSL_CTX_set_default_verify_paths(tls->ssl_ctx) != 1)
+            {
+                *errstr = xasprintf(_("cannot set X509 system trust for TLS session: %s"),
+                        ERR_error_string(ERR_get_error(), NULL));
+                SSL_CTX_free(tls->ssl_ctx);
+                tls->ssl_ctx = NULL;
+                return TLS_EFILE;
+            }
+        }
+        else
+        {
+            if (SSL_CTX_load_verify_locations(tls->ssl_ctx, trust_file, NULL) != 1)
+            {
+                *errstr = xasprintf(_("cannot load trust file %s: %s"),
+                        trust_file, ERR_error_string(ERR_get_error(), NULL));
+                SSL_CTX_free(tls->ssl_ctx);
+                tls->ssl_ctx = NULL;
+                return TLS_EFILE;
+            }
         }
         tls->have_trust_file = 1;
     }
-    if (sha256_fingerprint)
+    if (sha256_fingerprint && !no_certcheck)
     {
         memcpy(tls->fingerprint, sha256_fingerprint, 32);
         tls->have_sha256_fingerprint = 1;
     }
-    else if (sha1_fingerprint)
+    else if (sha1_fingerprint && !no_certcheck)
     {
         memcpy(tls->fingerprint, sha1_fingerprint, 20);
         tls->have_sha1_fingerprint = 1;
     }
-    else if (md5_fingerprint)
+    else if (md5_fingerprint && !no_certcheck)
     {
         memcpy(tls->fingerprint, md5_fingerprint, 16);
         tls->have_md5_fingerprint = 1;
@@ -1327,6 +1484,8 @@ int tls_init(tls_t *tls,
         tls->ssl_ctx = NULL;
         return TLS_ELIBFAILED;
     }
+    tls->no_certcheck = no_certcheck;
+    tls->hostname = xstrdup(hostname);
     return TLS_EOK;
 
 #endif /* HAVE_LIBSSL */
@@ -1363,7 +1522,7 @@ char *openssl_io_error(int error_code, int error_code2,
             {
                 if (error_code == 0)
                 {
-                    error_reason = _("a protocol violating EOF occured");
+                    error_reason = _("a protocol violating EOF occurred");
                 }
                 else if (error_code == -1)
                 {
@@ -1409,13 +1568,18 @@ char *openssl_io_error(int error_code, int error_code2,
  * see tls.h
  */
 
-int tls_start(tls_t *tls, int fd, const char *hostname, int no_certcheck,
-        tls_cert_info_t *tci, char **errstr)
+int tls_start(tls_t *tls, int fd,
+        tls_cert_info_t *tci, char **tls_parameter_description, char **errstr)
 {
 #ifdef HAVE_LIBGNUTLS
     int error_code;
 
+    gnutls_server_name_set(tls->session, GNUTLS_NAME_DNS, tls->hostname, strlen(tls->hostname));
+#if GNUTLS_VERSION_NUMBER < 0x030200
     gnutls_transport_set_ptr(tls->session, (gnutls_transport_ptr_t)fd);
+#else
+    gnutls_transport_set_int(tls->session, fd);
+#endif
     do
     {
         error_code = gnutls_handshake(tls->session);
@@ -1439,9 +1603,13 @@ int tls_start(tls_t *tls, int fd, const char *hostname, int no_certcheck,
             return error_code;
         }
     }
-    if (!no_certcheck)
+    if (tls_parameter_description)
     {
-        if ((error_code = tls_check_cert(tls, hostname, errstr)) != TLS_EOK)
+        *tls_parameter_description = gnutls_session_get_desc(tls->session);
+    }
+    if (!tls->no_certcheck)
+    {
+        if ((error_code = tls_check_cert(tls, errstr)) != TLS_EOK)
         {
             gnutls_deinit(tls->session);
             gnutls_certificate_free_credentials(tls->cred);
@@ -1491,9 +1659,13 @@ int tls_start(tls_t *tls, int fd, const char *hostname, int no_certcheck,
             return error_code;
         }
     }
-    if (!no_certcheck)
+    if (tls_parameter_description)
     {
-        if ((error_code = tls_check_cert(tls, hostname, errstr)) != TLS_EOK)
+        *tls_parameter_description = NULL; /* TODO */
+    }
+    if (!tls->no_certcheck)
+    {
+        if ((error_code = tls_check_cert(tls, errstr)) != TLS_EOK)
         {
             SSL_free(tls->ssl);
             SSL_CTX_free(tls->ssl_ctx);
@@ -1521,21 +1693,18 @@ int tls_readbuf_read(tls_t *tls, readbuf_t *readbuf, char *ptr,
 
     if (readbuf->count <= 0)
     {
-        ret = (int)gnutls_record_recv(tls->session,
-                readbuf->buf, sizeof(readbuf->buf));
+        do
+        {
+            ret = gnutls_record_recv(tls->session,
+                    readbuf->buf, sizeof(readbuf->buf));
+        }
+        while (ret == GNUTLS_E_AGAIN);
         if (ret < 0)
         {
             if (ret == GNUTLS_E_INTERRUPTED)
             {
                 *errstr = xasprintf(_("operation aborted"));
             }
-            else if (ret == GNUTLS_E_AGAIN)
-            {
-                /* This error message makes more sense than what
-                 * gnutls_strerror() would return. */
-                *errstr = xasprintf(_("cannot read from TLS connection: %s"),
-                        _("the operation timed out"));
-            }
             else
             {
                 *errstr = xasprintf(_("cannot read from TLS connection: %s"),
@@ -1655,19 +1824,17 @@ int tls_puts(tls_t *tls, const char *s, size_t len, char **errstr)
         return TLS_EOK;
     }
 
-    if ((ret = gnutls_record_send(tls->session, s, len)) < 0)
+    do
+    {
+        ret = gnutls_record_send(tls->session, s, len);
+    }
+    while (ret == GNUTLS_E_AGAIN);
+    if (ret < 0)
     {
         if (ret == GNUTLS_E_INTERRUPTED)
         {
             *errstr = xasprintf(_("operation aborted"));
         }
-        else if (ret == GNUTLS_E_AGAIN)
-        {
-            /* This error message makes more sense than what
-             * gnutls_strerror() would return. */
-            *errstr = xasprintf(_("cannot write to TLS connection: %s"),
-                    _("the operation timed out"));
-        }
         else
         {
             *errstr = xasprintf(_("cannot write to TLS connection: %s"),
@@ -1733,7 +1900,12 @@ void tls_close(tls_t *tls)
     if (tls->is_active)
     {
 #ifdef HAVE_LIBGNUTLS
-        gnutls_bye(tls->session, GNUTLS_SHUT_WR);
+        int e;
+        do
+        {
+            e = gnutls_bye(tls->session, GNUTLS_SHUT_WR);
+        }
+        while (e == GNUTLS_E_AGAIN);
         gnutls_deinit(tls->session);
         gnutls_certificate_free_credentials(tls->cred);
 #endif /* HAVE_LIBGNUTLS */
@@ -1743,6 +1915,10 @@ void tls_close(tls_t *tls)
         SSL_CTX_free(tls->ssl_ctx);
 #endif /* HAVE_LIBSSL */
     }
+    if (tls->hostname)
+    {
+        free(tls->hostname);
+    }
     tls_clear(tls);
 }
 
@@ -1759,3 +1935,30 @@ void tls_lib_deinit(void)
     gnutls_global_deinit();
 #endif /* HAVE_LIBGNUTLS */
 }
+
+
+/*
+ * tls_exitcode()
+ *
+ * see tls.h
+ */
+
+int tls_exitcode(int tls_error_code)
+{
+    switch (tls_error_code)
+    {
+        case TLS_EIO:
+            return EX_IOERR;
+        case TLS_EFILE:
+            return EX_NOINPUT;
+        case TLS_EHANDSHAKE:
+            return EX_PROTOCOL;
+        case TLS_ECERT:
+            /* did not find anything better... */
+            return EX_UNAVAILABLE;
+        case TLS_ELIBFAILED:
+        case TLS_ESEED:
+        default:
+            return EX_SOFTWARE;
+    }
+}
